local TWIRL_FRAMES = 8                -- Number of frames in the twirl (SPR2_PROP) animation.
local TWIRL_LENGTH = 25               -- The amount of frames a twirl lasts after being released.
local TWIRL_HOLD_WEIGHT = 8*FRACUNIT  -- While holding a twirl, Sparks will move down by this much each frame.
local TWIRL_MOMZ_BOOST = 20*FRACUNIT  -- When releasing a twirl, Sparks' momz will become this + her momz before starting the twirl.
local DROP_MOMZ = 25*FRACUNIT         -- The speed of the drop attack.
local DROP_DUST_SPAWN_RADIUS = 4
local DROP_DUST_SPEED = 4
local DROP_ATTACK_RADIUS = 128*FRACUNIT

local TWIRL_WITHOUT_GLIDE = true		-- If true, pushing Spin in mid-air when not gliding will do a twirl attack if possible, or
                                        -- a drop attack if not. If false, pushing Spin when gliding will always do a drop attack.
										-- Pushing Spin while gliding will never do a drop attack no matter what this is set to.
										-- I think False is better because it makes the behavior of Spin have a consistent effect,
										-- and makes the drop attack always accessible, but I've left this here in case I change my
										-- mind, or you want to change it yourself. I might make this configurable via a console
										-- command in a later version.
										-- ...is what I was *going* to say, but then I discovered that it becomes really easy to spam
										-- 1-frame "drop" attacks by holding Spin and jumping. Not sure if this is exploitable or not,
										-- but it seems rather messy.

local TWIRL_HOLD_SOUND = sfx_s3ka0      -- The sound to play while holding a twirl. (sfx_s25e also works)
local TWIRL_HOLD_SOUND_VOL = 255        -- The volume at which to play the hold sound. (128 works for sfx_s25e)
local TWIRL_HOLD_SOUND_TIME = 15        -- How many frames to wait before restarting the hold sound. (30 works for sfx_s25e)
local TWIRL_RELEASE_SOUND = sfx_drill1  -- The sound to play upon releasing a twirl.
local DROP_START_SOUND = sfx_trfire
local DROP_GROUND_SOUND = sfx_pstop     -- The sound to play upon hitting the ground at the end of a drop attack.

local GLIDE_UNDERWATER_START_SOUND = sfx_splash
local GLIDE_UNDERWATER_LOOP = sfx_amwtr6
local GLIDE_UNDERWATER_LOOP_VOL = 70

-- player.twirl_state:
--  0: Sparks is not twirling, but has the attack ready.
--  1: Sparks is currently holding a twirl (or has *just* released it)
--  2..TWIRL_LENGTH: Sparks has released a twirl, and is still spinning.
--  TWIRL_LENGTH+1..: Sparks has already used her twirl.

-- player.twirl_frame:
--  The number of frames since the player has last started a twirl. Increments even while holding it.

-- player.drop_state:
--  0: Sparks is not using her drop attack.
--  1: Sparks is using her drop attack.
--  2: Sparks is unable to use her drop attack because she was hit.

local function should_override_jump_spin(player)
	if player.rings >= 50 and All7Emeralds(emeralds) then
		-- Don't get in the way of a Super transformation
		return false
	elseif player.pflags & PF_THOKKED then
		-- This is so shields won't make it impossible to do a drop attack.
		return true
	else
		local count_as_no_shield = {SH_NONE, SH_PITY, SH_WHIRLWIND, SH_PINK}
		local shield_check = false
		for i,v in ipairs(count_as_no_shield) do
			if player.powers[pw_shield] == v then
				shield_check = true
			end
		end
		return shield_check
	end
end

local function spin_button_starts_twirl(player)
	-- Condition for whether or not a twirl should start upon pressing the spin button.
	if player.twirl_state ~= 0 then
		return false
	elseif player.secondjump then
		return true
	elseif P_IsObjectOnGround(player.mo) then
		return false
	else
		return should_override_jump_spin(player)
	end
end

local function start_drop(player)
	if player.drop_state == 0 then
		player.drop_state = 1
		S_StartSound(player.mo, DROP_START_SOUND)
	end
end

local function start_twirl(player)
	if player.twirl_state == 0 then
		player.twirl_frame = 0
	end
	player.twirl_state = 1
	if player.saved_momz == nil then
		-- Save our current momentum to add onto upon releasing
		player.saved_momz = player.mo.momz
	end
end

local function sparkle(player)
	for i=1,2 do
		local h_radius = 16
		local v_extent = 32
		local x = P_RandomRange(-h_radius, h_radius) * FRACUNIT
		local y = P_RandomRange(-h_radius, h_radius) * FRACUNIT
		local z = P_RandomRange(0, v_extent) * FRACUNIT
		local mo = P_SpawnMobjFromMobj(player.mo, x, y, z, MT_SPARKSPINSPARK)
		mo.color = player.mo.color
		mo.fuse = 2*TICRATE
	end
end

addHook("ThinkFrame", function()
	for player in players.iterate do
		if player.mo.skin ~= 'sparks' then
			continue
		end
		
		if player.twirl_state == nil then
			-- Initialize twirl_state
			player.twirl_state = 0
			player.gliding_frames = 0
		elseif player.twirl_state > 1 or (player.twirl_state == 1 and not (player.cmd.buttons & BT_SPIN)) then
			-- If we're twirling, and the player has already released the button, then increment twirl_state
			player.twirl_state = $+1
		end
		-- Increment twirl_frame
		player.twirl_frame = ($ or 0) + 1
		
		if player.twirl_state == 1 then  -- If we are currently holding a twirl...
			-- ...move the player toward the ground faster, since Sparks isn't supposed to be able to glide very far while attacking.
			player.mo.z = $-TWIRL_HOLD_WEIGHT*P_MobjFlip(player.mo)
		else
			S_StopSoundByID(player.mo, TWIRL_HOLD_SOUND) -- Stop the hold sound
			--if (player.cmd.buttons & BT_SPIN) and not (player.cmd.buttons & BT_JUMP) and player.twirl_state then
				-- If we've already released a twirl, and the player has released jump and pressed spin, trigger the drop move.
				-- moved to JumpSpinSpecial
				--start_drop(player)
			--end
		end
		
		if P_IsObjectOnGround(player.mo) then
			if player.drop_state == 1 then
				-- If we were dropping, and we've hit the ground, finish the move.
				S_StartSound(player.mo, DROP_GROUND_SOUND)
				for i=0,17 do
					local a = i*ANG1*20
					local x = cos(a)
					local y = sin(a)
					local dust = P_SpawnMobjFromMobj(player.mo, x*DROP_DUST_SPAWN_RADIUS, y*DROP_DUST_SPAWN_RADIUS, 4*FRACUNIT, MT_DUST)
					dust.color = player.mo.color
					dust.colorized = true
					dust.momx = DROP_DUST_SPEED * x
					dust.momy = DROP_DUST_SPEED * y
				end
				P_RadiusAttack(player.mo, player.mo, DROP_ATTACK_RADIUS)
			end
			-- Reset drop_state since we're on the ground, whether it's 1 or 2.
			player.drop_state = 0
		end
		
		if P_IsObjectInGoop(player.mo) or (player.mo.eflags & MFE_SPRUNG) then
			-- Abort the drop attack if we're in goop or hit a spring; otherwise the player would get stuck
			-- since P_IsObjectOnGround would always return false.
			player.drop_state = 0
		end
		
		-- Player state implementations
		if player.drop_state == 1 then
			-- Drop attack
			player.mo.sprite2 = SPR2_DROP
			player.mo.anim_duration = 1
			player.mo.frame = leveltime % 12
			player.mo.momz = -DROP_MOMZ*P_MobjFlip(player.mo)
			player.saved_momz = player.mo.momz
			--P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SUPERSPARK)
			sparkle(player)
		elseif player.twirl_state > 0 and player.twirl_state <= TWIRL_LENGTH and not P_IsObjectOnGround(player.mo) then
			-- Twirl attack
			player.mo.sprite2 = SPR2_PROP
			player.mo.frame = player.twirl_frame % TWIRL_FRAMES
			player.mo.anim_duration = 1
			if player.twirl_state == 1 and player.twirl_frame % TWIRL_HOLD_SOUND_TIME == 1 then
				-- Still holding
				S_StartSoundAtVolume(player.mo, TWIRL_HOLD_SOUND, TWIRL_HOLD_SOUND_VOL)
			elseif player.twirl_state == 2 then
				-- Just released
				S_StartSound(player.mo, TWIRL_RELEASE_SOUND)
				player.mo.momz = max(0, player.saved_momz) + TWIRL_MOMZ_BOOST*P_MobjFlip(player.mo)
				player.saved_momz = nil
			end
			sparkle(player)
		elseif player.dashmode > 3*TICRATE then
			-- Dash mode
			if player.mo.sprite2 ~= SPR2_DASH then
				-- Make sure it's always using the dash mode sprite
				player.mo.frame = $ % 2
			end
			player.mo.sprite2 = SPR2_DASH
			if player.secondjump then
				-- If we're gliding at the same time, leave a trail of glitter
				P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_NIGHTSPARKLE)
			end
		elseif player.secondjump then
			-- If we're gliding, and none of the above cases apply, then make sure it's using the float sprite.
			if player.mo.eflags & MFE_UNDERWATER then
				player.mo.sprite2 = SPR2_SWIM
				player.mo.frame = leveltime % 16
				player.mo.anim_duration = 8
				if player.gliding_frames % 2*TICRATE == 0 then
					if player.gliding_frames == 0 then
						S_StartSound(player.mo, GLIDE_UNDERWATER_START_SOUND)
					end
					if not S_SoundPlaying(player.mo, GLIDE_UNDERWATER_LOOP) then
						S_StartSoundAtVolume(player.mo, GLIDE_UNDERWATER_LOOP, GLIDE_UNDERWATER_LOOP_VOL)
					end
				end
			else
				-- Normally, the run animation would override it; this code will make sure that doesn't happen.
				-- Only displaying the first frame in that case is intentional, though it wasn't at first.
				player.mo.frame = $ % 2
				player.mo.sprite2 = SPR2_FLT_
			end
			player.gliding_frames = $ + 1
		end
		
		if player.pflags & PF_SPINNING or player.powers[pw_super] then
			sparkle(player)
		end
		
		if not player.secondjump then
			player.gliding_frames = 0
			S_StopSoundByID(player.mo, GLIDE_UNDERWATER_LOOP)
		end
		
		if P_IsObjectOnGround(player.mo) then
			-- If we're on the ground, reset the twirl attack.
			player.twirl_state = 0
			player.allow_another_twirl = false
		elseif (player.cmd.buttons & BT_SPIN) and player.secondjump and spin_button_starts_twirl(player) then
			-- Start the twirl attack if appropriate.
			-- Due to the "player.secondjump" check, this will only take effect while gliding, and is here because
			-- "JumpSpinSpecial" does not trigger while gliding. The code to start/hold a twirl when *not* gliding
			-- (not possible if TWIRL_WITHOUT_GLIDE is false) is in the "JumpSpinSpecial" handler below.
			start_twirl(player)
		elseif (player.allow_another_twirl or player.powers[pw_super]) and player.twirl_state > TWIRL_LENGTH then
			-- We've finished a twirl attack. It's normally only once per jump, so we don't usually reset it until hitting the ground.
			-- However, there are two conditions in which we do want the player to be able to twirl again:
			--  1) The player has hit an enemy in mid-air (which sets the "allow_another_twirl" flag), or
			--  2) The player is in Super form (in which case it's unlimited.)
			-- So, in these cases, we'll reset it.
			player.allow_another_twirl = false
			player.twirl_state = 0
		end
	end
end)

addHook("PlayerCanDamage", function(player)
	if player.valid and player.mo.skin == 'sparks'  -- If this is a valid player, playing as Sparks...
	and ((player.twirl_state > 0 and player.twirl_state <= TWIRL_LENGTH) or player.drop_state==1) then  -- ...and the player is either twirling or dropping...
		return true   -- ...then Sparks will attack enemies.
	end
	-- Otherwise, the function will return nil and the normal game logic will apply.
end)

local function mobjIsSparks(mo)
	-- Is the given mobj Sparks' player object?
	if mo == nil then
		return false
	else
		return mo.player ~= nil and mo.skin == 'sparks'
	end
end

addHook("MobjDamage", function(target, inflictor)
	if mobjIsSparks(target) then
		-- If Sparks *was* damaged, then prevent her from twirling or dropping until she reaches the ground.
		target.player.twirl_state = TWIRL_LENGTH + 1
		target.player.drop_state = 2
	end
	if mobjIsSparks(inflictor) then
		-- If Sparks was the one who *dealt* the damage, then set the allow_another_twirl flag.
		-- No need to check whether she's on the ground, as the flag will just be immediately cleared in that case.
		inflictor.player.allow_another_twirl = true
	end
end)

addHook("ShouldDamage", function(target, inflictor, source, damage, damagetype)
	if mobjIsSparks(target)
	 and inflictor == nil and source == nil and damagetype == DMG_FIRE
	 and target.player.dashmode > 3*TICRATE then
		-- Make Sparks immune to lava while in dash mode, because riding her scarf-board over lava is awesome.
		-- Plus it makes for interesting gameplay: if she slows down on lava, she'll exit dash mode and take damage.
		return false
	end
end)

addHook("JumpSpinSpecial", function(player)
	if player.mo.skin == 'sparks' then
		if TWIRL_WITHOUT_GLIDE then
			if spin_button_starts_twirl(player) then
				start_twirl(player)
				return true
			elseif player.twirl_state > 1 and should_override_jump_spin(player) then
				start_drop(player)
				return true
			end
		elseif should_override_jump_spin(player) and player.twirl_state ~= 1 then
			start_drop(player)
			return true
		end
	end
end)

addHook("AbilitySpecial", function(player)
	-- Prevent glide while dropping
	return player.mo.skin == 'sparks' and player.drop_state == 1
end)